#****************************************************************
# File: ./archiwe/store/read/archive.py
# (c) 2010, http://hakmatak.org, MIT License, ALL RIGHTS RESERVED
#****************************************************************
import os
import sys
import md5
import cPickle as pickle

from hakmatak.constant import ATTRS,LEAVES,NODES
from hakmatak.constant import DATA
from hakmatak.constant import NAME,VALUE

from hakmatak.node import Node
from hakmatak.leaf import Leaf

from hakmatak.store import Reader

from archiwe.constant import APP_NAME

import getpass
# cache is disabled
CACHE_DIR = None
# cache is enabled
#CACHE_DIR = os.path.join("/tmp","%s-%s"%(APP_NAME,getpass.getuser()),"archive")

# attribute names common to store of archive type
SIZE = "size"
TYPE = "type"
MTIME= "mtime"

class Archive(Reader):

    # can handle compressed file internally
    # override Reader.canDecompress
    canDecompress = True

    def __init__(self,path):
        Reader.__init__(self,path)

        # all meta info of this archive file
        cacheDir = CACHE_DIR
        #cacheDir = "/tmp"
        self.meta = self._get_all_meta(cacheDir=cacheDir)
        #import simplejson as json; print json.dumps(self.meta,indent=4)

    #....................
    # initialize all meta

    def _get_all_meta(self,cacheDir=None):
        # obtain all meta for this archive file.
        # if cacheDir is None, it is retrieved directly from file itself.
        # if cacheDir is not None:
        #   if cachePath exists and is newer, it is obtained from cachePath,
        #   otherwise obtain it from archive file and have it cached too.

        meta = None

        cachePath = None
        if cacheDir != None:
            tag = md5.new(self.path).hexdigest()
            cachePath = os.path.join(cacheDir,"%s.meta.pkl" % tag)

        # try to get from cachePath
        if cachePath != None and os.path.exists(cachePath) and os.stat(cachePath).st_mtime > os.stat(self.path).st_mtime:
            f = open(cachePath,"rb")
            meta = pickle.load(f)
            f.close()
            #print >> sys.stderr, "from cache"
            return meta

        # retrieve directly from archive file if cachePath non-existent or older
        meta = self._retrieve_all_meta_from_archive()
        #print >> sys.stderr, "from archive file"

        if cachePath == None:
            return meta
        # save to cachePath
        try:
            f = open(cachePath,"wb")
            pickle.dump(meta,f,-1)
            f.close()
        except Exception, e:
            if f != None:
                f.close()
            #os.remove(cachePath)
            raise Exception("%s. Failed to cache meta. Please remove file %s manually." % (e,cachePath))
        #print >> sys.stderr, "cache it"

        return meta

    # this should be overriden by subclass
    def _retrieve_all_meta_from_archive(self):
        raise Exception("%s is not defined" % sys._getframe().f_code.co_name)

    #...............
    # deal with meta

    def _get_meta_top(self,traverse=False):
        node = Node("")

        # if traverse
        if traverse == True:
            self._rearrange_nodes_from_dict_to_list(self.meta)
            if self.meta.has_key(LEAVES):
                node.leaves = self.meta[LEAVES]
            node.nodes = self.meta[NODES]
            return node.get_meta()

        # if not traverse
        # leaves
        if self.meta.has_key(LEAVES):
            self._rearrange_leaves_from_dict_to_list(self.meta)
            node.leaves = self.meta[LEAVES]
        # sub nodes
        for name,value in self.meta[NODES].iteritems():
            if ATTRS in value.keys():
                node.nodes.append({NAME:name,ATTRS:value[ATTRS]})
            else:
                node.nodes.append({NAME:name})
        return node.get_meta()

    def _get_meta_member(self,name,traverse=False):

        d = self.meta

        a = name[1:].split("/") # skip leading '/' and split
        parents = a[:-1] # they are all dirs
        member = a[-1] # it can be dir or a file

        # go to the immediate parent of the member while sanity checking
        for x in parents:
            if not d[NODES].has_key(x):
                raise Exception("Non-existent member %s" % name)
            d = d[NODES][x]

        # if it is a node
        if d[NODES].has_key(member):
            d = d[NODES][member]
            node = Node(member)
            if d.has_key(ATTRS):
                node.attrs = d[ATTRS]
            # if traverse
            if traverse == True:
                self._rearrange_nodes_from_dict_to_list(d)
                if d.has_key(LEAVES):
                    node.leaves = d[LEAVES]
                node.nodes = d[NODES]
                return node.get_meta()
            # if not traverse
            # leaves
            if d.has_key(LEAVES):
                self._rearrange_leaves_from_dict_to_list(d)
                node.leaves = d[LEAVES]
            # nodes
            keys = d[NODES].keys(); keys.sort()
            for key in keys: 
            #for name,value in d[NODES].iteritems():
                name = key
                value = d[NODES][key]
                if value.has_key(ATTRS):
                    node.nodes.append({NAME:name,ATTRS:value[ATTRS]})
                else:
                    node.nodes.append({NAME:name})
            return node.get_meta()

        # if it is a leaf
        if d[LEAVES].has_key(member):
            d = d[LEAVES][member]
            leaf = Leaf(member)
            leaf.attrs = d[ATTRS]
            return leaf.get_meta()

        raise Exception("Non-existent member %s" % name)

    def get_meta(self,name,traverse=False):
        if name == '':
            return self._get_meta_top(traverse=traverse)
        return self._get_meta_member(name,traverse=traverse)

    #...............
    # deal with data

    # this should be overriden by subclass
    def _get_data_member(self,name):
        raise Exception("%s is not defined" % sys._getframe().f_code.co_name)

    def get_data(self,name,indexer=None,flatten=False):
        # remove leading '/' to court zip convention
        fname = name[1:]
        return {DATA:self._get_data_member(fname)}

    #.............................
    # below are supporting methods

    def _rearrange_nodes_from_dict_to_list(self,d):
        self._rearrange_leaves_from_dict_to_list(d)
        if NODES not in d:
            return
        l = []
        keys = d[NODES].keys(); keys.sort()
        for key in keys:
            value = d[NODES][key]
            # recursively go down to subnodes
            if NODES in value:
                self._rearrange_nodes_from_dict_to_list(value)
            value[NAME] = key
            l.append(value)
        # remove empty nodes
        if l == []:
            del d[NODES]
        else:
            d[NODES] = l

    def _rearrange_leaves_from_dict_to_list(self,d):
        if LEAVES not in d:
            return
        l = []
        keys = d[LEAVES].keys(); keys.sort
        for key in keys:
            value = d[LEAVES][key]
            value[NAME] = key
            l.append(value)
        # remove empty leaves
        if l == []:
            del d[LEAVES]
        else:
            d[LEAVES] = l



#****************************************************************
# File: ./archiwe/store/write/tar.py
# (c) 2010, http://hakmatak.org, MIT License, ALL RIGHTS RESERVED
#****************************************************************
from hakmatak.store import Writer

import tarfile
import StringIO

import os

class StoreWriter(Writer):

    def __init__(self,path):
        # make sure path points to a tar file if it exists
        if os.path.isfile(path):
            if os.path.getsize(path) == 0:
                raise Exception("File %s has zero size." % path)
            if not tarfile.is_tarfile(path):
                raise Exception("Not a tar file %s" % path)
        # superclass init
        Writer.__init__(self,path)

    def put_data(self,name,indexer=None,data=None):
        # currently no support for append/update for data slice
        if indexer not in [None,""]:
            raise Exception("put_data is not supported for indexer %s." % indexer)

        if data == None:
            raise Exception("Data is None.")

        # a member
        tarPath = name[1:] # strip leading '/'
        tarInfo = tarfile.TarInfo(name=tarPath)
        # ref: http://stackoverflow.com/questions/740820/python-write-string-directly-to-tarfile
        tarInfo.size = len(data)

        fileObj = StringIO.StringIO(data)

        tarFile = None
        if os.path.isfile(self.path):
            tarFile = tarfile.open(name=self.path,mode="a")
        else:
            tarFile = tarfile.open(name=self.path,mode="w")
        tarFile.addfile(tarInfo,fileobj=fileObj)
        tarFile.close()

        return {"status":"success"}



#****************************************************************
# File: ./archiwe/store/read/tar.py
# (c) 2010, http://hakmatak.org, MIT License, ALL RIGHTS RESERVED
#****************************************************************
import os
from datetime import datetime

import tarfile

from hakmatak.constant import ATTRS,NODES,LEAVES
from hakmatak.constant import NAME,VALUE
from hakmatak.constant import DATA

from archive import SIZE,TYPE,MTIME
from archive import Archive

def timestamp_to_isoformat(ts):
    return datetime.fromtimestamp(ts).isoformat('T')

class StoreReader(Archive):

    # override Archive._retrieve_all_meta_from_archive()
    def _retrieve_all_meta_from_archive(self):
        #tar = tarfile.open(self.path,"r:*")
        tar = tarfile.open(self.path,"r")
        d = {}
        for tarinfo in tar:
            name = tarinfo.name
            #print name
            l = name.split('/')
            # this list must have at least 2 members
            if len(l) < 2:
                raise Exception('wrong entry in tar')
            # if a node is found
            if l[-1] == '':
                if NODES not in d:
                    d[NODES] = {}
                parent = d
                nodes = d[NODES]
                for x in l[:-1]:
                    if x not in nodes:
                        nodes[x] = {}
                    if NODES not in nodes[x]:
                        nodes[x][NODES] = {}
                    parent = nodes[x]
                    nodes = nodes[x][NODES]
                parent[ATTRS] = [
                    {NAME:SIZE,VALUE:tarinfo.size},
                    #{NAME:TYPE,VALUE:tarinfo.type},
                    {NAME:MTIME,VALUE:timestamp_to_isoformat(tarinfo.mtime)}
                    ]
            # if a file is found
            else:
                if NODES not in d:
                    d[NODES] = {}
                parent = d
                nodes = d[NODES]
                for x in l[:-1]:
                    if x not in nodes:
                        nodes[x] = {}
                    if NODES not in nodes[x]:
                        nodes[x][NODES] = {}
                    parent = nodes[x]
                    nodes = nodes[x][NODES]
                if LEAVES not in parent:
                    parent[LEAVES] = {}
                if l[-1] not in parent[LEAVES]:
                    parent[LEAVES][l[-1]] = {}
                    parent[LEAVES][l[-1]][ATTRS] = [
                        {NAME:SIZE,VALUE:tarinfo.size},
                        #{NAME:TYPE,VALUE:tarinfo.type},
                        {NAME:MTIME,VALUE:timestamp_to_isoformat(tarinfo.mtime)}
                        ]
        tar.close()

        return d

    # override Archive._get_data_member()
    def _get_data_member(self,name):
        #tar = tarfile.open(self.path,"r:*")
        tar = tarfile.open(self.path,"r")
        fobj = tar.extractfile(name)
        bytes = fobj.read()
        fobj.close()
        tar.close()
        return bytes



#****************************************************************
# File: ./archiwe/store/read/factory.py
# (c) 2010, http://hakmatak.org, MIT License, ALL RIGHTS RESERVED
#****************************************************************
from hakmatak.store import ReaderClassFactory as ParentFactory

class ReaderClassFactory(ParentFactory):

    def create_reader_class(self,w10nType=None):
        if w10nType == "tar":
            from tar import StoreReader
            return StoreReader
        if w10nType == "zip":
            from zip import StoreReader
            return StoreReader
        raise Exception("Unable to create reader class for w10n type %s." % w10nType)
        #return ParentFactory.create_reader_class(self,w10nType=w10nType)

def main():

    readerClassFactory = ReaderClassFactory()

    for w10nType in ["tar","zip"]:
        readerClass = readerClassFactory.create_reader_class(w10nType)
        print readerClass

if __name__ == "__main__":
    main()



#****************************************************************
# File: ./archiwe/store/read/zip.py
# (c) 2010, http://hakmatak.org, MIT License, ALL RIGHTS RESERVED
#****************************************************************
import os
from datetime import datetime

import zipfile

from hakmatak.constant import ATTRS,NODES,LEAVES
from hakmatak.constant import NAME,VALUE
from hakmatak.constant import DATA

from archive import SIZE,TYPE,MTIME
from archive import Archive

def date_time_to_isoformat(date_time):
    # http://docs.python.org/library/zipfile.html
    # m and d are one based, H,M,S are zero based.
    y,m,d,H,M,S = date_time
    return datetime(y,m,d,H,M,S).isoformat('T')

class StoreReader(Archive):

    # override Archive._retrieve_all_meta_from_archive()
    def _retrieve_all_meta_from_archive(self):
        zip = zipfile.ZipFile(self.path,"r")
        d = {}
        for zipinfo in zip.infolist():
            name = zipinfo.filename
            # do not handle such a case now!
            if name[0] == '/':
                raise Exception("can't handle name with leading '/': %s"%name)
            # WARNING: this is our hack to make sure
            # variable 'name' is an adequate path, suitable for processing
            # by algorithm below, which counts on '/' to work properly
            if name[0:2] != './':
                name = './' + name
            #print name
            l = name.split('/')
            #print l
            # this list must have at least 2 members
            if len(l) < 2:
                raise Exception('wrong entry in zip: %s'%name)
            # a node is found
            if l[-1] == '':
                if NODES not in d:
                    d[NODES] = {}
                parent = d
                nodes = d[NODES]
                for x in l[:-1]:
                    if x not in nodes:
                        nodes[x] = {}
                    if NODES not in nodes[x]:
                        nodes[x][NODES] = {}
                    parent = nodes[x]
                    nodes = nodes[x][NODES]
                parent[ATTRS] = [
                    {NAME:SIZE,VALUE:zipinfo.file_size},
                    # in zip, there is no concept of dir?
                    #{NAME:TYPE,VALUE:zipinfo.type}, 
                    {NAME:MTIME,VALUE:date_time_to_isoformat(zipinfo.date_time)}
                    ]
            # a leaf is found
            else:
                if NODES not in d:
                    d[NODES] = {}
                parent = d
                nodes = d[NODES]
                for x in l[:-1]:
                    if x not in nodes:
                        nodes[x] = {}
                    if NODES not in nodes[x]:
                        nodes[x][NODES] = {}
                    parent = nodes[x]
                    nodes = nodes[x][NODES]
                if LEAVES not in parent:
                    parent[LEAVES] = {}
                if l[-1] not in parent[LEAVES]:
                    parent[LEAVES][l[-1]] = {}
                    parent[LEAVES][l[-1]][ATTRS] = [
                        {NAME:SIZE,VALUE:zipinfo.file_size},
                        # no type concept?
                        #{NAME:TYPE,VALUE:zipinfo.type},
                        {NAME:MTIME,VALUE:date_time_to_isoformat(zipinfo.date_time)}
                        ]
        zip.close()

        #import simplejson as json; print json.dumps(d,indent=4)

        # internal sanity checks consistent with our hack above
        if NODES not in d or ('.' not in d[NODES].keys()):
            raise Exception("internal error")

        # remove one extra level introduced by our hack above
        d = d[NODES]['.']

        #import simplejson as json; print json.dumps(d,indent=4)

        d[NAME] = ''

        # we might not need this
        #if NODES not in d:
        #    d[NODES] = []

        ## one more sanity check consistent with our hack above
        #if ATTRS in d:
        #    raise Exception("internal error")

        return d

    # override Archive._get_data_member()
    def _get_data_member(self,name):
        zip = zipfile.ZipFile(self.path,"r")
        bytes = zip.read(name)
        zip.close()
        return bytes



#****************************************************************
# File: ./archiwe/store/write/zip.py
# (c) 2010, http://hakmatak.org, MIT License, ALL RIGHTS RESERVED
#****************************************************************
from hakmatak.store import Writer

import zipfile

import os

class StoreWriter(Writer):

    def __init__(self,path):
        # make sure path points to a zip file if it exists
        if os.path.isfile(path) and not zipfile.is_zipfile(path):
            raise Exception("Not a zip file %s" % path)
        # superclass init
        Writer.__init__(self,path)

    def put_data(self,name,indexer=None,data=None):
        # currently no support for append/update for data slice
        if indexer not in [None,""]:
            raise Exception("put_data is not supported for indexer %s." % indexer)
        
        if data == None:
            raise Exception("Data is None.")

        zipFile = None
        if os.path.isfile(self.path):
            zipFile = zipfile.ZipFile(self.path,"a")
        else:
            zipFile = zipfile.ZipFile(self.path,"w")
        zipPath = name[1:] # strip leading '/'
        zipFile.writestr(zipPath,data)
        zipFile.close()

        return {"status":"success"}



#****************************************************************
# File: ./archiwe/store/write/factory.py
# (c) 2010, http://hakmatak.org, MIT License, ALL RIGHTS RESERVED
#****************************************************************
from hakmatak.store import WriterClassFactory as ParentFactory

class WriterClassFactory(ParentFactory):
    def create_writer_class(self,w10nType=None):
        if w10nType == "tar":
            from tar import StoreWriter
            return StoreWriter
        if w10nType == "zip":
            from zip import StoreWriter
            return StoreWriter
        raise Exception("Unable to create writer class for w10n type %s." % w10nType)

def main():

    writerClassFactory = WriterClassFactory()

    for w10nType in ["tar","zip"]:
        writerClass = writerClassFactory.create_writer_class(w10nType)
        print writerClass

if __name__ == "__main__":
    main()
